#!/usr/local/bin/python

# GLTools PpenGL Basics Wrapping by Javi Agenjo '09 javi.agenjo@gmail.com

from OpenGL.GL import *
from OpenGL.GLU import *
from OpenGL.GLUT import *
from OpenGL.GL.ARB.framebuffer_object import *
from OpenGL.GL.EXT.framebuffer_object import *

from PIL import Image

import pygame
import Image
import sys,time

from ctypes import *
from math import *

def createAndCompileShader(type,source):
        shader=glCreateShader(type)
        glShaderSource(shader,source)
        glCompileShader(shader)

        # get "compile status" - glCompileShader will not fail with
        # an exception in case of syntax errors

        result=glGetShaderiv(shader,GL_COMPILE_STATUS)

        if (result!=1): # shader didn't compile
                raise Exception("Couldn't compile shader\nShader compilation Log:\n"+glGetShaderInfoLog(shader))
        return shader

def enableRenderHUD():
	glPushAttrib( GL_ENABLE_BIT )

	glDisable( GL_DEPTH_TEST )
	glDisable( GL_CULL_FACE )
	glDisable( GL_TEXTURE_2D )
	
	glMatrixMode( GL_MODELVIEW )
	glPushMatrix()
	glLoadIdentity()
	glMatrixMode( GL_PROJECTION )
	glPushMatrix()
	glLoadIdentity()
	gluOrtho2D(0,1,0,1)

def disableRenderHUD():
	glMatrixMode( GL_PROJECTION )
	glPopMatrix()
	glMatrixMode( GL_MODELVIEW )
	glPopMatrix()

	glPopAttrib()
	

class Shader:
    def __init__(self):
        print "Error: parameters needed"
        
    def __init__(self, vs_source, ps_source):
        
        vs_shader = createAndCompileShader(GL_VERTEX_SHADER,vs_source)
        ps_shader = createAndCompileShader(GL_FRAGMENT_SHADER,ps_source)

        self.program = glCreateProgram()
        glAttachShader(self.program,vs_shader)
        glAttachShader(self.program,ps_shader)
        glLinkProgram(self.program)

        self.enable()
        self.disable()
        self.last_slot = 0
        
    def setUniform1f(self,name,value):
        loc = glGetUniformLocation(self.program,name);
        if (loc == -1):
            print "Error: "+name+" var not found in shader"
        else:
            glUniform1f(loc,value);

    def setUniform2f(self,name,value):
        loc = glGetUniformLocation(self.program,name);
        if (loc == -1):
            print "Error: "+name+" var not found in shader"
        else:
            glUniform2f(loc,value[0],value[1]);


    def setUniformTexture(self,name,tex_id):
        loc = glGetUniformLocation(self.program,name);
        if (loc == -1):
            print "Error: "+name+" var not found in shader"
        else:
                glActiveTexture( GL_TEXTURE0 + self.last_slot )
                glBindTexture( GL_TEXTURE_2D, tex_id )
                glUniform1i(loc,self.last_slot)
                self.last_slot += 1

    def enable(self):
        try:
                glUseProgram(self.program)
                self.last_slot = 0
        except OpenGL.error.GLError:
                print glGetProgramInfoLog(self.program)
                raise

    def disable(self):
            glUseProgram(0)

# ******************************************************

class Texture:
    """Params: mipmaps (contains mipmaps), mag_filter, min_filter, wrap"""

    def generic_init(self,parameters):
        #global data
        self.ready = False
        self.parameters = {}
        self.parameters['mipmaps'] = True
        self.parameters['mag_filter'] = GL_LINEAR
        self.parameters['min_filter'] = GL_LINEAR_MIPMAP_LINEAR        
        self.parameters['wrap'] = GL_REPEAT
        self.parameters.update(parameters)
        
    def __init__(self,parameters={}):
        self.generic_init(parameters)
            
    def load(self,filename):

        try:
                im = Image.open(filename)
        except IOError:
                print "File " + filename + " could not be opened"
                self.ready = False
                return False
                
        self.width, self.height = im.size
        self.mode = im.mode
        self.channels = len(im.getbands())

        #data = im.tostring("raw", "RGBX", 0, -1)
        data = im.tostring()

        types = [GL_NONE, GL_LUMINANCE, GL_NONE, GL_RGB, GL_RGBA]
        self.depth = types[ self.channels ]
        #print "Channels " + str(self.channels)
                
        # generate texture
        if not hasattr(self,"texture_id"):
                self.texture_id=glGenTextures(1)

        self.updateParameters()
        glPixelStorei(GL_UNPACK_ALIGNMENT,1)
        if self.parameters['mipmaps'] == True:
                gluBuild2DMipmaps(GL_TEXTURE_2D, self.depth, self.width, self.height, GL_RGBA, GL_UNSIGNED_BYTE, data)
        else:
                glTexImage2D(GL_TEXTURE_2D, 0, self.depth, self.width, self.height, 0, GL_RGBA, GL_UNSIGNED_BYTE, data)
        self.ready = True
        #print len ( im.tostring() )
        return True

    def free(self):
        if hasattr(self,"texture_id"):
                glDeleteTextures(self.texture_id)
                delattr(self,texture_id)

    def bind(self):
        if not hasattr(self,"texture_id"):
                return
        glBindTexture( GL_TEXTURE_2D, self.texture_id )
        glEnable( GL_TEXTURE_2D )

    def unbind(self):
        glDisable( GL_TEXTURE_2D )        
                
    def render(self, x=0,y=0,width=1,height=1):
        if not hasattr(self,"texture_id"):
                return
        enableRenderHUD()
        self.bind()

        glBegin(GL_QUADS)
        glColor3f(1,1,1)
        glTexCoord2f(0,0)
        glVertex3f(x,y,0)
        glTexCoord2f(1,0)
        glVertex3f(x+width,y,0)
        glTexCoord2f(1,1)
        glVertex3f(x+width,y+height,0)
        glTexCoord2f(0,1)
        glVertex3f(x,y+height,0)
        glEnd()

        disableRenderHUD()
	
    def updateParameters(self):
        self.bind()
        glTexParameterf( GL_TEXTURE_2D, GL_TEXTURE_WRAP_S,self.parameters['wrap'])
        glTexParameterf( GL_TEXTURE_2D, GL_TEXTURE_WRAP_T,self.parameters['wrap'])
        glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, self.parameters['mag_filter'])
        glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, self.parameters['min_filter'])
        

class RenderTexture(Texture):
    def __init__(self,width,height,depth,parameters={}):

        if not 'wrap' in parameters: parameters['wrap'] = GL_CLAMP
        Texture.__init__(self,parameters)
        
        if ( bool(glGenFramebuffers) == False):
                print "This graphics card do not support FBOs"
                self.ready = False
                return

        self.texture_id = glGenTextures( 1 )
        self.width = width
        self.height = height
        self.depth = depth

        glBindTexture( GL_TEXTURE_2D, self.texture_id )
        glTexEnvf( GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_MODULATE )
        self.updateParameters()
        glTexImage2D(GL_TEXTURE_2D, 0,depth,width,height,0,depth,GL_UNSIGNED_BYTE, None)

        if self.parameters['mipmaps']:
            glGenerateMipmap(GL_TEXTURE_2D)

        self.fbo = glGenFramebuffers(1)
        self.ready = True

    def enable(self):
        if not hasattr(self,"fbo"):
            return
        glBindFramebufferEXT(GL_FRAMEBUFFER_EXT, self.fbo)
	# In case of errors or suspect behaviour, try this:
	#    print "Framebuffer Status:"
	#    print glCheckFramebufferStatusEXT(GL_FRAMEBUFFER_EXT);
	
        glFramebufferTexture2DEXT(GL_FRAMEBUFFER_EXT, GL_COLOR_ATTACHMENT0_EXT,
        GL_TEXTURE_2D, self.texture_id, 0)
        glPushAttrib(GL_VIEWPORT_BIT) # save viewport
        glViewport(0, 0, self.width, self.height) # set to fire effect dimensions

    def disable(self):
        if not hasattr(self,"fbo"):
            return
        glPopAttrib()
        glBindFramebufferEXT(GL_FRAMEBUFFER_EXT, 0)
        if self.parameters['mipmaps']:
            glBindTexture( GL_TEXTURE_2D, self.texture_id )
            glGenerateMipmap(GL_TEXTURE_2D)

    def clear(self):
        self.enable()
        glClearColor(1.0, 0.0, 0.0, 0.0)
        glClear(GL_COLOR_BUFFER_BIT|GL_DEPTH_BUFFER_BIT)
        self.disable()


class Primitive:
    def __init__(self,type):

        self.cubelist = glGenLists(1);

        glNewList(self.cubelist,GL_COMPILE)

        if (type == "cube"):
            glBegin(GL_QUADS)
            glNormal3f(0,0,-1)
            glTexCoord2f(0,0), glVertex3f( -1, 1, -1)
            glTexCoord2f(1,0), glVertex3f( 1, 1, -1)
            glTexCoord2f(1,1), glVertex3f( 1, -1, -1)
            glTexCoord2f(0,1), glVertex3f( -1, -1, -1)
            
            glNormal3f(0,0,1)
            glTexCoord2f(1,1), glVertex3f( -1, -1, 1)
            glTexCoord2f(0,1), glVertex3f( 1, -1, 1)
            glTexCoord2f(0,0), glVertex3f( 1, 1, 1)
            glTexCoord2f(1,0), glVertex3f( -1, 1, 1)

            glNormal3f(0,-1,0)
            glTexCoord2f(0,1), glVertex3f( -1, -1, -1)
            glTexCoord2f(1,1), glVertex3f( 1, -1, -1)
            glTexCoord2f(1,0), glVertex3f( 1, -1, 1)
            glTexCoord2f(0,0), glVertex3f( -1, -1, 1)

            glNormal3f(0,1,0)
            glTexCoord2f(1,0), glVertex3f( -1, 1, 1)
            glTexCoord2f(0,0), glVertex3f( 1, 1, 1)
            glTexCoord2f(0,1), glVertex3f( 1, 1, -1)
            glTexCoord2f(1,1), glVertex3f( -1, 1, -1)

            glNormal3f(-1,0,0)
            glTexCoord2f(0,1), glVertex3f( -1, -1, 1)
            glTexCoord2f(0,0), glVertex3f( -1, 1, 1)
            glTexCoord2f(1,0), glVertex3f( -1, 1, -1)
            glTexCoord2f(1,1), glVertex3f( -1, -1, -1)

            glNormal3f(1,0,0)
            glTexCoord2f(0,1), glVertex3f( 1, -1, -1)
            glTexCoord2f(0,0), glVertex3f( 1, 1, -1)
            glTexCoord2f(1,0), glVertex3f( 1, 1, 1)
            glTexCoord2f(1,1), glVertex3f( 1, -1, 1)
            glEnd()
            
        glEndList()

    def render(self):
        glCallList(self.cubelist)

# *************************************
def renderFPSbar(elapsed_time_in_ms):
    if (elapsed_time_in_ms == 0): elapsed_time = 1
    total = 0.01666 / (elapsed_time_in_ms * 0.001)

    enableRenderHUD()

    glDisable( GL_BLEND )
    glLineWidth(4)
    glBegin(GL_LINES)
    glColor3f(0.5,0.5,0.7)
    glVertex3f(-1,-1,0)
    glVertex3f(-1 + 2 * total,-1,0)
    glEnd()

    disableRenderHUD()

def distance(a,b):
    v = [ b[0]-a[0], b[1]-a[1], b[2]-a[2] ]
    return sqrt(v[0]*v[0]+v[1]*v[1]+v[2]*v[2])

def jpg_file_write(name, number, data):
    im = Image.frombuffer("RGBA", (800,600), data, "raw", "RGBA", 0, 0)
    fnumber = "%05d" % number
    im.save(name + fnumber + ".jpg")
